import {
  MatcherException,
  MatcherExceptionCode,
} from '../../../exceptions/matcher.exception';
import { IRequestGeography } from '../../interfaces/geography-request.interface';
import { PointType } from '../../../../geography/domain/types/point-type.enum';
import { Point } from '../../../../geography/domain/types/point.type';
import { Route } from './route';
import { Role } from '../../types/role.enum';
import { IGeorouter } from '../../interfaces/georouter.interface';
import { Waypoint } from './waypoint';
import { Actor } from './actor';
import { Person } from './person';
import { Step } from '../../types/step.enum';
import { Path } from '../../types/path.type';
import { IFindTimezone } from '../../../../geography/domain/interfaces/timezone-finder.interface';
import { Timezoner } from './timezoner';

export class Geography {
  private geographyRequest: IRequestGeography;
  private person: Person;
  private points: Point[];
  originType: PointType;
  destinationType: PointType;
  timezones: string[];
  driverRoute: Route;
  passengerRoute: Route;
  timezoneFinder: IFindTimezone;

  constructor(
    geographyRequest: IRequestGeography,
    timezoner: Timezoner,
    person: Person,
  ) {
    this.geographyRequest = geographyRequest;
    this.person = person;
    this.points = [];
    this.originType = undefined;
    this.destinationType = undefined;
    this.timezones = [timezoner.timezone];
    this.timezoneFinder = timezoner.finder;
  }

  init = (): void => {
    this.validateWaypoints();
    this.setTimezones();
    this.setPointTypes();
  };

  createRoutes = async (
    roles: Role[],
    georouter: IGeorouter,
  ): Promise<void> => {
    let driverWaypoints: Waypoint[] = [];
    let passengerWaypoints: Waypoint[] = [];
    const paths: Path[] = [];
    if (roles.includes(Role.DRIVER) && roles.includes(Role.PASSENGER)) {
      if (this.points.length == 2) {
        // 2 points => same route for driver and passenger
        const commonPath: Path = {
          key: RouteKey.COMMON,
          points: this.points,
        };
        driverWaypoints = this.createWaypoints(commonPath.points, Role.DRIVER);
        passengerWaypoints = this.createWaypoints(
          commonPath.points,
          Role.PASSENGER,
        );
        paths.push(commonPath);
      } else {
        const driverPath: Path = {
          key: RouteKey.DRIVER,
          points: this.points,
        };
        driverWaypoints = this.createWaypoints(driverPath.points, Role.DRIVER);
        const passengerPath: Path = {
          key: RouteKey.PASSENGER,
          points: [this.points[0], this.points[this.points.length - 1]],
        };
        passengerWaypoints = this.createWaypoints(
          passengerPath.points,
          Role.PASSENGER,
        );
        paths.push(driverPath, passengerPath);
      }
    } else if (roles.includes(Role.DRIVER)) {
      const driverPath: Path = {
        key: RouteKey.DRIVER,
        points: this.points,
      };
      driverWaypoints = this.createWaypoints(driverPath.points, Role.DRIVER);
      paths.push(driverPath);
    } else if (roles.includes(Role.PASSENGER)) {
      const passengerPath: Path = {
        key: RouteKey.PASSENGER,
        points: [this.points[0], this.points[this.points.length - 1]],
      };
      passengerWaypoints = this.createWaypoints(
        passengerPath.points,
        Role.PASSENGER,
      );
      paths.push(passengerPath);
    }
    const routes = await georouter.route(paths, {
      withDistance: false,
      withPoints: true,
      withTime: false,
    });
    if (routes.some((route) => route.key == RouteKey.COMMON)) {
      this.driverRoute = routes.find(
        (route) => route.key == RouteKey.COMMON,
      ).route;
      this.passengerRoute = routes.find(
        (route) => route.key == RouteKey.COMMON,
      ).route;
      this.driverRoute.setWaypoints(driverWaypoints);
      this.passengerRoute.setWaypoints(passengerWaypoints);
    } else {
      if (routes.some((route) => route.key == RouteKey.DRIVER)) {
        this.driverRoute = routes.find(
          (route) => route.key == RouteKey.DRIVER,
        ).route;
        this.driverRoute.setWaypoints(driverWaypoints);
      }
      if (routes.some((route) => route.key == RouteKey.PASSENGER)) {
        this.passengerRoute = routes.find(
          (route) => route.key == RouteKey.PASSENGER,
        ).route;
        this.passengerRoute.setWaypoints(passengerWaypoints);
      }
    }
  };

  private validateWaypoints = (): void => {
    if (this.geographyRequest.waypoints.length < 2) {
      throw new MatcherException(
        MatcherExceptionCode.INVALID_ARGUMENT,
        'At least 2 waypoints are required',
      );
    }
    this.geographyRequest.waypoints.map((point) => {
      if (!this.isValidPoint(point)) {
        throw new MatcherException(
          MatcherExceptionCode.INVALID_ARGUMENT,
          `Waypoint { Lon: ${point.lon}, Lat: ${point.lat} } is not valid`,
        );
      }
      this.points.push(point);
    });
  };

  private setTimezones = (): void => {
    this.timezones = this.timezoneFinder.timezones(
      this.geographyRequest.waypoints[0].lat,
      this.geographyRequest.waypoints[0].lon,
    );
  };

  private setPointTypes = (): void => {
    this.originType =
      this.geographyRequest.waypoints[0].type ?? PointType.OTHER;
    this.destinationType =
      this.geographyRequest.waypoints[
        this.geographyRequest.waypoints.length - 1
      ].type ?? PointType.OTHER;
  };

  private isValidPoint = (point: Point): boolean =>
    this.isValidLongitude(point.lon) && this.isValidLatitude(point.lat);

  private isValidLongitude = (longitude: number): boolean =>
    longitude >= -180 && longitude <= 180;

  private isValidLatitude = (latitude: number): boolean =>
    latitude >= -90 && latitude <= 90;

  private createWaypoints = (points: Point[], role: Role): Waypoint[] => {
    return points.map((point, index) => {
      const waypoint = new Waypoint(point);
      if (index == 0) {
        waypoint.addActor(new Actor(this.person, role, Step.START));
      } else if (index == points.length - 1) {
        waypoint.addActor(new Actor(this.person, role, Step.FINISH));
      } else {
        waypoint.addActor(new Actor(this.person, role, Step.INTERMEDIATE));
      }
      return waypoint;
    });
  };
}

export enum RouteKey {
  COMMON = 'common',
  DRIVER = 'driver',
  PASSENGER = 'passenger',
}
