import { MatchRequest } from '../domain/dtos/match.request';
import { Geography } from '../domain/entities/ecosystem/geography';
import { Person } from '../domain/entities/ecosystem/person';
import { Requirement } from '../domain/entities/ecosystem/requirement';
import { Role } from '../domain/types/role.enum';
import { AlgorithmSettings } from '../domain/entities/ecosystem/algorithm-settings';
import { Time } from '../domain/entities/ecosystem/time';
import { IDefaultParams } from '../domain/types/default-params.type';
import { IGeorouter } from '../domain/interfaces/georouter.interface';
import { ICreateGeorouter } from '../domain/interfaces/georouter-creator.interface';
import { IFindTimezone } from '../../geography/domain/interfaces/timezone-finder.interface';

export class MatchQuery {
  private readonly matchRequest: MatchRequest;
  private readonly defaultParams: IDefaultParams;
  private readonly georouterCreator: ICreateGeorouter;
  person: Person;
  roles: Role[];
  time: Time;
  geography: Geography;
  exclusions: number[];
  requirement: Requirement;
  algorithmSettings: AlgorithmSettings;
  georouter: IGeorouter;
  timezoneFinder: IFindTimezone;

  constructor(
    matchRequest: MatchRequest,
    defaultParams: IDefaultParams,
    georouterCreator: ICreateGeorouter,
    timezoneFinder: IFindTimezone,
  ) {
    this.matchRequest = matchRequest;
    this.defaultParams = defaultParams;
    this.georouterCreator = georouterCreator;
    this.timezoneFinder = timezoneFinder;
    this.setPerson();
    this.setRoles();
    this.setTime();
    this.setGeography();
    this.setRequirement();
    this.setAlgorithmSettings();
    this.setExclusions();
  }

  createRoutes = (): void => {
    this.geography.createRoutes(this.roles, this.algorithmSettings.georouter);
  };

  private setPerson = (): void => {
    this.person = new Person(
      this.matchRequest,
      this.defaultParams.DEFAULT_IDENTIFIER,
      this.defaultParams.MARGIN_DURATION,
    );
    this.person.init();
  };

  private setRoles = (): void => {
    this.roles = [];
    if (this.matchRequest.driver) this.roles.push(Role.DRIVER);
    if (this.matchRequest.passenger) this.roles.push(Role.PASSENGER);
    if (this.roles.length == 0) this.roles.push(Role.PASSENGER);
  };

  private setTime = (): void => {
    this.time = new Time(
      this.matchRequest,
      this.defaultParams.MARGIN_DURATION,
      this.defaultParams.VALIDITY_DURATION,
    );
    this.time.init();
  };

  private setGeography = (): void => {
    this.geography = new Geography(
      this.matchRequest,
      {
        timezone: this.defaultParams.DEFAULT_TIMEZONE,
        finder: this.timezoneFinder,
      },
      this.person,
    );
    this.geography.init();
  };

  private setRequirement = (): void => {
    this.requirement = new Requirement(
      this.matchRequest,
      this.defaultParams.DEFAULT_SEATS,
    );
  };

  private setAlgorithmSettings = (): void => {
    this.algorithmSettings = new AlgorithmSettings(
      this.matchRequest,
      this.defaultParams.DEFAULT_ALGORITHM_SETTINGS,
      this.time.frequency,
      this.georouterCreator,
    );
  };

  private setExclusions = (): void => {
    this.exclusions = [];
    if (this.matchRequest.identifier)
      this.exclusions.push(this.matchRequest.identifier);
    if (this.matchRequest.exclusions)
      this.exclusions.push(...this.matchRequest.exclusions);
  };
}
